package com.spiffymap.naviupdate.model;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.Optional;

/**
 * 23-May-2017
 *
 * @author steve
 */
public class Directory extends Entry implements Serializable {

    private static final long serialVersionUID = -4263774999225147322L;
    private final static char DIRECTORY = 'D';
    private final static char FILE = 'F';

    private final List<Entry> entries = new ArrayList<>();

    public Directory() {
        super();
    }

    public Directory(String name, Object parent, List<Entry> entries) {
        super(name, parent);
        this.entries.addAll(entries);
    }

    public Directory(String name, Object parent) {
        super(name, parent);
    }

    /**
     * Return all file entries in the directory tree.
     *
     * @return set of file entries
     */
    public Set<FileMetaDataEntry> getAllFileEntries() {
        Set<FileMetaDataEntry> result = new HashSet<>();
        for (Entry entry : entries) {
            if (entry instanceof FileMetaDataEntry) {
                result.add((FileMetaDataEntry) entry);
            } else {
                // Directory
                Directory directory = (Directory) entry;
                Set<FileMetaDataEntry> fileEntries = directory.getAllFileEntries();
                result.addAll(fileEntries);
            }
        }
        return result;
    }

    public Set<Directory> immediateSubdirectories() {
        Set<Directory> result = new HashSet<>();
        for (Entry entry : entries) {
            if (entry instanceof Directory) {
                result.add((Directory) entry);
            }
        }
        return result;
    }

    public Optional<Directory> findImmediateSubDirectoryWithName(String name) {
        for (Entry entry : entries) {
            if (entry instanceof Directory) {
                if (entry.getName().equalsIgnoreCase(name)) {
                    return Optional.of((Directory) entry);
                }
            }
        }
        return Optional.empty();
    }

    /**
     *
     * @return true if this directory has at least one subdirectory.
     */
    public boolean hasSubDirectories() {
        return entries.stream().anyMatch(entry -> entry instanceof Directory);
    }

    /**
     * @return the entries
     */
    public List<Entry> getEntries() {
        return entries;
    }

    /**
     * Root directory has tree as parent.
     *
     * @return true if root
     */
    public boolean isRoot() {
        return getParent() == null;
    }

    /**
     * Get path relative to file tree.
     *
     * @return path
     */
    @Override
    public Path getRelativePath() {
        if (isRoot()) {
            return Paths.get("");
        } else {
            Path parentPath = ((Directory) getParent()).getRelativePath();
            return parentPath.resolve(getName());
        }
    }

    private void saveCompressedEntryName(Entry entry, DataOutput out) throws IOException {
        byte[] nameBytes = entry.getName().getBytes(StandardCharsets.UTF_8);
        out.writeByte(nameBytes.length);
        out.write(nameBytes);
    }

    private String loadCompressedString(DataInput in) throws IOException {
        int size = (in.readByte() & 0xFF);
        byte[] bytes = new byte[size];
        in.readFully(bytes);
        return new String(bytes, StandardCharsets.UTF_8);
    }

    private void saveCompressedFile(FileMetaDataEntry entry, DataOutput out) throws IOException {
        out.writeByte(FILE);
        saveCompressedEntryName(entry, out);
        out.writeLong(entry.getSize());
        byte[] hash = entry.getHash();
        out.writeByte(hash.length);
        out.write(hash);
    }

    private FileMetaDataEntry loadCompressedFile(DataInput in) throws IOException {
        String name = loadCompressedString(in);
        long size = in.readLong();
        int hashSize = in.readByte() & 0xFF;
        byte[] hash = new byte[hashSize];
        in.readFully(hash);
        return new FileMetaDataEntry();
    }

    public void saveCompressed(DataOutput out) throws IOException {
        out.writeByte(DIRECTORY);
        short entriesSize = (short) entries.size();
        out.writeShort(entriesSize);
        byte[] nameBytes = getName().getBytes(StandardCharsets.UTF_8);
        out.writeByte(nameBytes.length);
        out.write(nameBytes);
        for (Entry entry : getEntries()) {

        }

    }

    /**
     * Return all entries which are directories with no subdirectories within
     * this directory.
     *
     * @return list of all terminal directories
     */
    public Set<Directory> getAllTerminalDirectories() {
        Set<Directory> terminalDirectories = new HashSet<>();
        for (Entry entry : entries) {
            if (entry instanceof Directory) {
                Directory subDirectory = (Directory) entry;
                if (!subDirectory.hasSubDirectories()) {
                    terminalDirectories.add(subDirectory);
                } else {
                    Set<Directory> subTerminalDirectories = subDirectory.getAllTerminalDirectories();
                    terminalDirectories.addAll(subTerminalDirectories);
                }
            }
        }
        return terminalDirectories;
    }

    /**
     *
     * @return the file tree.
     */
    public FileTree getTree() {
        if (isRoot()) {
            return (FileTree) getParent();
        } else {
            return ((Directory) getParent()).getTree();
        }
    }

    /**
     *
     * @return the relative path with Unix separators.
     */
    public String relativePathUnixSeparators() {
        Path relativePath = getRelativePath();
        StringBuilder pathName = new StringBuilder();
        boolean first = true;
        Iterator<Path> paths = relativePath.iterator();
        while (paths.hasNext()) {
            Path subPath = paths.next();
            if (!first) {
                pathName.append('/');
            }
            pathName.append(subPath.getFileName());
            first = false;
        }
        return pathName.toString();
    }

    public void add(Entry entry) {
        entries.add(entry);
    }

    public void addAll(Collection<Entry> newEntries) {
        entries.addAll(newEntries);
    }

    /**
     * @param entries the entries to set
     */
    public void setEntries(List<Entry> entries) {
        this.entries.clear();
        this.entries.addAll(entries);
    }
}
