package com.spiffymap.naviupdate.model;

import com.spiffymap.util.shared.ValueHolder;
import com.spiffymap.util.shared.date.DayTimeStamp;
import java.io.File;
import java.io.PrintWriter;
import java.io.Serializable;
import java.io.StringWriter;
import java.math.BigInteger;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 *
 * @author steve
 */
public class FileTree implements Serializable {

    public static class EntryAndDepth {

        private final Entry entry;
        private final int depth;

        public EntryAndDepth(Entry entry, int depth) {
            this.entry = entry;
            this.depth = depth;
        }

        /**
         * @return the entry
         */
        public Entry getEntry() {
            return entry;
        }

        /**
         * @return the depth
         */
        public int getDepth() {
            return depth;
        }
    }

    private static final long serialVersionUID = -2376346613657122734L;

    private String repositoryName;
    private Directory rootEntry;
    private String timeStamp;
    private byte[] hash;

    public FileTree() {
        setTimeStamp(new DayTimeStamp());
    }

    /**
     * @return the rootEntry
     */
    public Directory getRootEntry() {
        return rootEntry;
    }

    /**
     * Find the highest directory with the supplied name
     *
     * @param name
     * @return directory or empty
     */
    public Optional<Directory> findDirectoryWithName(String name) {
        List<Directory> dirs = getEntriesMatching(entry -> (entry instanceof Directory) && entry.getName().equals(name)).stream()
                .map(entry -> (Directory) entry).collect(Collectors.toList());
        if (dirs.isEmpty()) {
            return Optional.empty();
        }
        Collections.sort(dirs, (dir1, dir2) -> new Integer(dir1.getName().length()).compareTo(dir2.getName().length()));
        return Optional.of(dirs.get(0));
    }

    /**
     * Find the entries matching a supplied predicate. The entries closest to
     * the root of the tree are returned.
     *
     * @param predicate
     * @return list of entries
     */
    public Set<Entry> getEntriesMatching(Predicate<Entry> predicate) {
        SortedMap<Integer, Set<Entry>> entryMap = new TreeMap<>();
        searchRecursive(rootEntry, predicate, entryMap, 0);
        if (entryMap.isEmpty()) {
            return new HashSet<>();
        } else {
            return entryMap.get(entryMap.firstKey());
        }
    }

    /**
     * Search for matching entries.
     *
     * @param directory
     * @param predicate
     * @param entryMap matching entry map, keyed on depth in tree.
     * @param depth depth in tree
     */
    private void searchRecursive(Directory directory, Predicate<Entry> predicate, Map<Integer, Set<Entry>> entryMap, int depth) {
        for (Entry entry : directory.getEntries()) {
            if (predicate.test(entry)) {
                Set<Entry> entrySet = entryMap.getOrDefault(depth, new HashSet<>());
                entrySet.add(entry);
                entryMap.put(depth, entrySet);
            } else {
                if (entry instanceof Directory) {
                    searchRecursive((Directory) entry, predicate, entryMap, depth + 1);
                }
            }
        }
    }

    /**
     * Run a consumer on all entries
     *
     * @param directory
     * @param consumer
     */
    public void consumeRecursive(Directory directory, Consumer<FileMetaDataEntry> consumer) {
        for (Entry entry : directory.getEntries()) {
            if (entry instanceof Directory) {
                consumeRecursive((Directory) entry, consumer);
            } else {
                FileMetaDataEntry fileEntry = (FileMetaDataEntry) entry;
                consumer.accept(fileEntry);
            }
        }
    }

    /**
     * @param entryName the repositoryName of the rootEntry to set
     */
    public void setRootEntry(String entryName) {
        rootEntry = new Directory(entryName, null);
    }

    /**
     * @param entry the root entry
     */
    public void setRootEntry(Directory entry) {
        rootEntry = entry;
    }

    /**
     * @return the repositoryName
     */
    public String getRepositoryName() {
        return repositoryName;
    }

    /**
     * @param repositoryName the repositoryName to set
     */
    public void setRepositoryName(String repositoryName) {
        this.repositoryName = repositoryName;
    }

    /**
     * @return the timeStamp
     */
    public DayTimeStamp getTimeStamp() {
        if (timeStamp == null) {
            return null;
        } else {
            return DayTimeStamp.fromString(timeStamp);
        }
    }

    /**
     * @param timeStamp the timeStamp to set
     */
    public final void setTimeStamp(DayTimeStamp timeStamp) {
        if (timeStamp == null) {
            this.timeStamp = null;
        } else {
            this.timeStamp = timeStamp.toString();
        }
    }

    public void clear() {
        rootEntry.getEntries().clear();
    }

    public Set<String> getAllRelativeFilePaths() {
        Set<Path> allRelativePaths = getAllRelativeFilePaths(getRootEntry());
        Set<String> allRelativePathStrings = new HashSet<>();
        for (Path path : allRelativePaths) {
            allRelativePathStrings.add(path.toString());
        }
        return allRelativePathStrings;
    }

    public int size() {
        return getAllRelativeFilePaths().size();
    }

    private Set<Path> getAllRelativeFilePaths(Directory directory) {
        // Set to check for duplicates.  e.g. 'This\File' and 'THIS\file'.
        Set<String> unixStylePathStrings = new HashSet<>();
        Set<Path> paths = new HashSet<>();
        for (Entry entry : directory.getEntries()) {
            if (entry instanceof Directory) {
                paths.addAll(getAllRelativeFilePaths((Directory) entry));
            } else {
                String unixStylePathString = entry.getUnixStyleRelativePathString();
                if (!unixStylePathStrings.contains(unixStylePathString)) {
                    paths.add(entry.getRelativePath());
                    unixStylePathStrings.add(unixStylePathString);
                } else {
                    throw new IllegalArgumentException("Duplicate: " + entry.getRelativePath());
                }
            }
        }
        return paths;
    }

    public Set<Directory> getDataDirectoriesForCellAndEdition(String cellName, int edition) {
        String cellNameLowerCase = cellName.toLowerCase();
        String source = cellNameLowerCase.substring(0, 2);
        // Find the cell directory
        Optional<Directory> optSourceDirectory = rootEntry.findImmediateSubDirectoryWithName(source);
        if (!optSourceDirectory.isPresent()) {
            return new HashSet<>();
        }
        Directory sourceDirectory = optSourceDirectory.get();
        Optional<Directory> optCellDirectory = sourceDirectory.findImmediateSubDirectoryWithName(cellName);
        if (!optCellDirectory.isPresent()) {
            return new HashSet<>();
        }
        Directory cellDirectory = optCellDirectory.get();
        String editionString = String.valueOf(edition);
        Optional<Directory> optEditionDirectory = cellDirectory.findImmediateSubDirectoryWithName(editionString);
        if (!optEditionDirectory.isPresent()) {
            return new HashSet<>();
        }
        Directory editionDirectory = optEditionDirectory.get();
        return editionDirectory.getAllTerminalDirectories();
    }

    /**
     * Get file paths as a single string for compactness.
     *
     * @return file paths string
     */
    public String getFilePaths() {
        List<FileMetaDataEntry> entries = new ArrayList<>();
        consumeRecursive(rootEntry, entry -> {
            if (entry instanceof FileMetaDataEntry) {
                entries.add((FileMetaDataEntry) entry);
            }
        });
        Collections.sort(entries, (entry1, entry2) -> entry1.getUnixStyleRelativePathString().
                compareTo(entry2.getUnixStyleRelativePathString()));
        StringWriter buffer = new StringWriter();
        try (PrintWriter out = new PrintWriter(buffer)) {
            for (FileMetaDataEntry entry : entries) {
                out.println(entry.getUnixStyleRelativePathString());
            }
        }
        return buffer.toString();
    }

    /**
     * The overall hash is calculated: it's the sum of all file paths hashes
     * with Unix separators and the file contents hashes.
     */
    public void calculateHash() {
        ValueHolder<BigInteger> sum = new ValueHolder<>();
        sum.setValue(BigInteger.ZERO);
        consumeRecursive(rootEntry, entry -> {
            if (entry instanceof FileMetaDataEntry) {
                try {
                    FileMetaDataEntry metaData = (FileMetaDataEntry) entry;
                    MessageDigest digest = MessageDigest.getInstance("MD5");
                    String unixPathLowerCase = metaData.getRelativePath().toString().toLowerCase().replace(File.separatorChar, '/');
                    digest.update(unixPathLowerCase.getBytes());
                    byte[] pathHash = digest.digest();
                    BigInteger pathHashInt = new BigInteger(pathHash);
                    BigInteger hashInt = new BigInteger(((FileMetaDataEntry) entry).getHash());
                    BigInteger hashSum = pathHashInt.add(hashInt);
                    BigInteger overallSum = sum.getValue().add(hashSum);
                    sum.setValue(overallSum);
                } catch (NoSuchAlgorithmException nsae) {
                    // Something is seriously wrong - throw a runtime exception
                    throw new RuntimeException(nsae);
                }
            }
        });
        hash = sum.getValue().toByteArray();
    }

    /**
     * @return the hash
     */
    public byte[] getHash() {
        return hash;
    }

    /**
     * @param hash the hash to set
     */
    public void setHash(byte[] hash) {
        this.hash = hash;
    }

    public Set<FileMetaDataEntry> getAllFileMetaData() {
        Set<FileMetaDataEntry> entries = rootEntry.getAllFileEntries();
        return entries;
    }

    /**
     * Return a copy of this object with only meta data
     *
     * @return meta data
     */
    public FileTree getMetaDataFileTree() {
        Directory metaDataDirectory = getMetaData(rootEntry, null);
        FileTree metaDataFileTree = new FileTree();
        metaDataFileTree.setRootEntry(metaDataDirectory);
        metaDataFileTree.setRepositoryName(repositoryName);
        metaDataFileTree.setTimeStamp(DayTimeStamp.fromString(timeStamp));
        return metaDataFileTree;
    }

    /**
     * Recursively collect metadata
     *
     * @param directory
     * @param metaDirectoryParent parent directory for metadata
     * @return directory containing metadata
     */
    private Directory getMetaData(Directory directory, Directory metaDirectoryParent) {
        List<Entry> entries = new ArrayList<>();
        Directory metaDataDirectory = new Directory();
        for (Entry entry : directory.getEntries()) {
            if (entry instanceof FileDataEntry) {
                FileDataEntry dataEntry = (FileDataEntry) entry;
                FileMetaDataEntry metaDataEntry = new FileMetaDataEntry(entry.getName(), directory,
                        dataEntry.getHash(), dataEntry.getData().length);
                entries.add(metaDataEntry);
            } else if (entry instanceof Directory) {
                Directory subDirectory = (Directory) entry;
                Directory metaDataSubDirectory = getMetaData(subDirectory, metaDataDirectory);
                entries.add(metaDataSubDirectory);
            }
        }
        metaDataDirectory.setEntries(entries);
        metaDataDirectory.setParent(metaDirectoryParent);
        return metaDataDirectory;
    }

    /**
     * Remove an entry from the tree.
     *
     * @param entry
     */
    public void removeEntry(Entry entry) {
        Object parent = entry.getParent();
        entry.setParent(null);
        Directory directory = (Directory) parent;
        directory.getEntries().remove(entry);
    }
}
