package com.spiffymap.naviupdate.model;

import com.spiffymap.util.server.FileUtils;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import org.springframework.util.StreamUtils;

/**
 *
 * @author steve
 */
public class Repository {

    public final static String CATALOG_FILE = "CATALOG.031";
    public final static String PRODUCTS_FILE = "PRODUCTS.TXT";
    public final static String VERSIONS_DIRECTORY = "version_history";
    public final static Pattern UPDATED_PRODUCT_FILE_PATTERN = Pattern.compile("\\d\\d\\d\\d\\d\\d\\.txt");
    private String name;
    private String mainPath;
    private String updatePath;
    private Date timestamp;
    private List<Supplier> suppliers;
    private FileTree fileTree;
    private FileTree updatesFileTree;
    private FileTree serverFileTree;
    private String encRootDirectory;
    private String encRootNameToUse;
    private String productsDirectory;
    private byte[] fileTreeHash;
    private byte[] updatesFileTreeHash;
    private String filePaths;

    public Repository(String name, String mainPath, String updatePath, Date timestamp, List<Supplier> suppliers) {
        this.name = name;
        if (mainPath == null) {
            throw new IllegalArgumentException("Path can't be null");
        }
        this.mainPath = mainPath;
        this.updatePath = updatePath;
        this.timestamp = timestamp;
        this.suppliers = suppliers;
    }

    public static String getRepositoryListFileName(String repositoryName) {
        return System.getProperty("java.io.tmpdir") + File.separator + repositoryName + ".lst";
    }

    @Override
    public int hashCode() {
        int hash = 5;
        hash = 67 * hash + Objects.hashCode(this.name);
        hash = 67 * hash + Objects.hashCode(this.timestamp);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Repository other = (Repository) obj;
        if (!Objects.equals(this.name, other.name)) {
            return false;
        }
        if (!Objects.equals(this.timestamp, other.timestamp)) {
            return false;
        }
        return true;
    }

    /**
     * Get the most recent products file, or empty if not found.
     *
     * @return contents of products file or empty
     * @throws IOException
     */
    public Optional<byte[]> getMostRecentProductsFile() throws IOException {
        // Check for updated versions of the products file
        File productsFileToUse = null;
        File versionsDirectory = Paths.get(mainPath).resolve(VERSIONS_DIRECTORY).toFile();
        if (versionsDirectory.exists() && versionsDirectory.isDirectory()) {
            File[] files = versionsDirectory.listFiles();
            List<File> filesList = new ArrayList<>();
            for (File file : files) {
                if (UPDATED_PRODUCT_FILE_PATTERN.matcher(file.getName()).matches()) {
                    filesList.add(file);
                }
            }
            if (!filesList.isEmpty()) {
                // Sort the files and get the most recent
                Collections.sort(filesList, (file1, file2) -> file1.getName().compareTo(file2.getName()));
                productsFileToUse = filesList.get(filesList.size() - 1);
            }
        }
        if (productsFileToUse == null) {
            // Use 'products.txt'
            Path productsDirectoryPath = Paths.get(mainPath).resolve(productsDirectory);
            Path productsPath = productsDirectoryPath.resolve(Repository.PRODUCTS_FILE.toUpperCase());
            if (!productsPath.toFile().exists()) {
                productsPath = productsDirectoryPath.resolve(Repository.PRODUCTS_FILE.toLowerCase());
                if (!productsPath.toFile().exists()) {
                    return Optional.empty();
                }
            }
            productsFileToUse = productsPath.toFile();
        }
        try (InputStream in = new FileInputStream(productsFileToUse)) {
            byte[] contents = StreamUtils.copyToByteArray(in);
            return Optional.of(contents);
        }
    }

    @Override
    public String toString() {
        return "Repository{" + "name=" + name + ", timestamp=" + timestamp + '}';
    }

    /**
     * @return the timestamp
     */
    public Date getTimestamp() {
        return timestamp;
    }

    /**
     * Return data for a cell or empty if not found.
     *
     * @param cellName
     * @return cell data or empty
     */
    public Optional<CellData> getDataLocationForCell(String cellName) {
        if (cellName.length() < 2) {
            return Optional.empty();
        }
        String supplierName = cellName.substring(0, 2);
        for (Supplier supplier : getSuppliers()) {
            if (supplier.getName().equals(supplierName)) {
                List<CellData> cellDataSet = supplier.getCellData();
                for (CellData cellData : cellDataSet) {
                    if (cellData.getName().equals(cellName)) {
                        return Optional.of(cellData);
                    }
                }
            }
        }
        return Optional.empty();
    }

    public Map<Path, byte[]> getCellData(CellData cellData) throws IOException {
        Path repositoryPath = Paths.get(mainPath);
        // Map of path to file contents
        Map<Path, byte[]> data = new HashMap<>();
        List<CellDataDirectory> cellDataDirectories = cellData.getCellDataDirectories();
        for (CellDataDirectory directory : cellDataDirectories) {
            Path relativePathToDirectory = Paths.get(directory.getRelativePath());
            Set<FileData> files = new HashSet<>();
            files.addAll(directory.getCellFiles());
            files.addAll(directory.getSignatureFiles());
            files.addAll(directory.getOtherFiles());
            for (FileData fileData : files) {
                Path relativePathToFile = relativePathToDirectory.resolve(fileData.getName());
                Path fullPathToFile = repositoryPath.resolve(relativePathToFile);
                byte[] fileContents = FileUtils.loadFileAsByteArray(fullPathToFile.toFile());
                data.put(relativePathToFile, fileContents);
            }
        }
        return data;
    }

    /**
     * Convert a string with path separator '/' (as from a URL) to a path
     *
     * @param pathString
     * @return path instance
     */
    public static Path getPathFromString(String pathString) {
        List<String> pathParts = Arrays.asList(pathString.split("/"));
        String osDependentPathString = pathParts.stream().collect(Collectors.joining(File.separator));
        return Paths.get(osDependentPathString);
    }

    /**
     * Return a listing of the supplied repository-relative directory, or
     * Optional empty if not found.
     *
     * @param directory
     * @return list or empty.
     * @throws IOException
     */
    public Optional<List<String>> listDirectory(String directory) throws IOException {
        Path relativePath = getPathFromString(directory);
        Path repositoryTopPath = Paths.get(getPath());
        Path fullPath = repositoryTopPath.resolve(relativePath).normalize().toAbsolutePath();
        // Check that the full path isn't outside of the repository
        if (!fullPath.startsWith(repositoryTopPath.toAbsolutePath())) {
            throw new IOException("Path is outside of repository");
        }
        File file = fullPath.toFile();
        if (!file.exists()) {
            return Optional.empty();
        }
        if (!file.isDirectory()) {
            throw new IOException("Supplied string does not represent a directory");
        }
        // Get the names of files in the directory
        File[] filesInDirectory = file.listFiles();
        List<String> fileNames = new ArrayList<>();
        for (File fileInDirectory : filesInDirectory) {
            String fileName = fileInDirectory.toPath().getFileName().toString();
            fileNames.add(fileName);
        }
        return Optional.of(fileNames);
    }

    /**
     * Return the contents of the supplied repository-relative file, or Optional
     * empty if not found.
     *
     * @param fileName
     * @return list or empty.
     * @throws IOException
     */
    public Optional<byte[]> fetch(String fileName) throws IOException {
        Path relativePath = getPathFromString(fileName);
        Path repositoryTopPath = Paths.get(getPath());
        Path fullPath = repositoryTopPath.resolve(relativePath).normalize().toAbsolutePath();
        // Check that the full path isn't outside of the repository
        if (!fullPath.startsWith(repositoryTopPath.toAbsolutePath())) {
            throw new IOException("Path is outside of repository");
        }
        File file = fullPath.toFile();
        if (!file.exists()) {
            return Optional.empty();
        }
        if (file.isDirectory()) {
            throw new IOException("Supplied string does not represent a file");
        }
        try (InputStream in = new FileInputStream(file)) {
            byte[] contents = StreamUtils.copyToByteArray(in);
            return Optional.of(contents);
        }
    }

    /**
     * Return bytes containing zipped entries from file data
     *
     * @param cellDataAndPaths
     * @return byte array
     * @throws IOException
     */
    public byte[] getZippedData(Map<Path, byte[]> cellDataAndPaths) throws IOException {
        ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        ZipOutputStream out = new ZipOutputStream(buffer);
        for (Path cellPath : cellDataAndPaths.keySet()) {
            byte[] data = cellDataAndPaths.get(cellPath);
            // Get path as a string containing Unix separators
            String filePath = cellPath.toString().replaceAll("\\\\", "/");
            ZipEntry entry = new ZipEntry(filePath);
            out.putNextEntry(entry);
            out.write(data);
            out.closeEntry();
        }
        out.close();
        buffer.close();
        return buffer.toByteArray();
    }

    /**
     * @return the name
     */
    public String getName() {
        return name;
    }

    /**
     * @return the path
     */
    public String getPath() {
        return mainPath;
    }

    /**
     * @return the fileTree
     */
    public FileTree getFileTree() {
        return fileTree;
    }

    /**
     * @param fileTree the fileTree to set
     */
    public void setFileTree(FileTree fileTree) {
        this.fileTree = fileTree;
    }

    /**
     * @return the encRootDirectory
     */
    public String getEncRootDirectory() {
        return encRootDirectory;
    }

    /**
     * @param encRootDirectory the encRootDirectory to set
     */
    public void setEncRootDirectory(String encRootDirectory) {
        this.encRootDirectory = encRootDirectory;
    }

    /**
     * @return the productsDirectory
     */
    public String getProductsDirectory() {
        return productsDirectory;
    }

    /**
     * @param productsDirectory the productsDirectory to set
     */
    public void setProductsDirectory(String productsDirectory) {
        this.productsDirectory = productsDirectory;
    }

    /**
     * @return the updatePath
     */
    public String getUpdatePath() {
        return updatePath;
    }

    /**
     * @return the fileTreeHash
     */
    public Optional<byte[]> getFileTreeHash() {
        return Optional.ofNullable(fileTreeHash);
    }

    /**
     * @return the updatesFileTreeHash
     */
    public Optional<byte[]> getUpdatesFileTreeHash() {
        return Optional.ofNullable(updatesFileTreeHash);
    }

    /**
     * @param fileTreeHash the fileTreeHash to set
     */
    public void setFileTreeHash(byte[] fileTreeHash) {
        this.fileTreeHash = fileTreeHash;
    }

    /**
     * @return the suppliers
     */
    public List<Supplier> getSuppliers() {
        return suppliers;
    }

    /**
     * @return the updatesFileTree
     */
    public FileTree getUpdatesFileTree() {
        return updatesFileTree;
    }

    /**
     * @param updatesFileTree the updatesFileTree to set
     */
    public void setUpdatesFileTree(FileTree updatesFileTree) {
        this.updatesFileTree = updatesFileTree;
    }

    /**
     * @return the serverFileTree
     */
    public FileTree getServerFileTree() {
        return serverFileTree;
    }

    /**
     * @param serverFileTree the serverFileTree to set
     */
    public void setServerFileTree(FileTree serverFileTree) {
        this.serverFileTree = serverFileTree;
    }

    /**
     * @return the encRootNameToUse
     */
    public String getEncRootNameToUse() {
        return encRootNameToUse;
    }

    /**
     * @param encRootNameToUse the encRootNameToUse to set
     */
    public void setEncRootNameToUse(String encRootNameToUse) {
        this.encRootNameToUse = encRootNameToUse;
    }

    /**
     * @param name the name to set
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @param mainPath the mainPath to set
     */
    public void setMainPath(String mainPath) {
        this.mainPath = mainPath;
    }

    /**
     * @param updatePath the updatePath to set
     */
    public void setUpdatePath(String updatePath) {
        this.updatePath = updatePath;
    }

    /**
     * @param timestamp the timestamp to set
     */
    public void setTimestamp(Date timestamp) {
        this.timestamp = timestamp;
    }

    /**
     * @param suppliers the suppliers to set
     */
    public void setSuppliers(List<Supplier> suppliers) {
        this.suppliers = suppliers;
    }

    /**
     * @return the filePaths
     */
    public String getFilePaths() {
        return filePaths;
    }

    /**
     * @param filePaths the filePaths to set
     */
    public void setFilePaths(String filePaths) {
        this.filePaths = filePaths;
    }

    /**
     * Find missing file paths from a supplied set. Paths
     *
     * @param paths
     * @return
     */
    public Set<String> findMissingFilePaths(Set<String> paths) {
        Set<String> standardisedPaths = paths.stream().
                map(path -> (encRootDirectory + File.separatorChar + path).toLowerCase().replace(File.separatorChar, '/')).
                collect(Collectors.toSet());
        try (BufferedReader in = new BufferedReader(new StringReader(filePaths))) {
            String line = in.readLine();
            while (line != null) {
                standardisedPaths.remove(line);
                if (standardisedPaths.isEmpty()) {
                    break;
                }
                line = in.readLine();
            }
            return standardisedPaths;
        } catch (IOException ioe) {
            // Should never happen
            throw new RuntimeException(ioe);
        }
    }

    public boolean hasUpdatePath() {
        return updatePath != null;
    }
}
